/* Copyright (C) The Scalable Software Infrastructure Project. All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
   3. Neither the name of the project nor the names of its contributors 
      may be used to endorse or promote products derived from this software 
      without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY THE SCALABLE SOFTWARE INFRASTRUCTURE PROJECT
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE SCALABLE SOFTWARE INFRASTRUCTURE
   PROJECT BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
*/

#ifdef HAVE_CONFIG_H
	#include "lis_config.h"
#else
#ifdef HAVE_CONFIG_WIN_H
	#include "lis_config_win.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_MALLOC_H
        #include <malloc.h>
#endif
#include <string.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_array_swap
 * lis_array_copy
 * lis_array_scale
 * lis_array_axpy
 * lis_array_xpay
 * lis_array_axpyz
 * lis_array_pmul
 * lis_array_pdiv
 * lis_array_set_all
 * lis_array_abs
 * lis_array_reciprocal
 * lis_array_shift
 * lis_array_dot
 * lis_array_dot2
 * lis_array_nrm2
 * lis_array_nrm1
 * lis_array_sum
 * lis_array_matvec
 * lis_array_matvect
 * lis_array_matvec2
 * lis_array_matinv
 * lis_array_LUdecomp
 * lis_array_invGauss
 * lis_array_invvec
 * lis_array_invvect
 * lis_array_matmat
 * lis_array_matmat2
 * lis_array_solve
 * lis_array_cgs
 * lis_array_mgs
 * lis_array_qr
 * lis_array_power
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_array_swap"
LIS_INT lis_array_swap(LIS_INT n, LIS_SCALAR *x, LIS_SCALAR *y)
{
  LIS_INT i;
  LIS_SCALAR t;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      t = y[i];
      y[i] = x[i];
      x[i] = t;
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_copy"
LIS_INT lis_array_copy(LIS_INT n, LIS_SCALAR *x, LIS_SCALAR *y)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      y[i] = x[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_scale"
LIS_INT lis_array_scale(LIS_INT n, LIS_SCALAR alpha, LIS_SCALAR *v)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      v[i] = alpha * v[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_axpy"
LIS_INT lis_array_axpy(LIS_INT n, LIS_SCALAR alpha, LIS_SCALAR *x, LIS_SCALAR *y)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      y[i] = alpha * x[i] + y[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_xpay"
LIS_INT lis_array_xpay(LIS_INT n, LIS_SCALAR alpha, LIS_SCALAR *x, LIS_SCALAR *y)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      y[i] = x[i] + alpha * y[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_axpyz"
LIS_INT lis_array_axpyz(LIS_INT n, LIS_SCALAR alpha, LIS_SCALAR *x, LIS_SCALAR *y, LIS_SCALAR *z)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      z[i] = alpha * x[i] + y[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_pmul"
LIS_INT lis_array_pmul(LIS_INT n, LIS_SCALAR *x, LIS_SCALAR *y, LIS_SCALAR *z)
{
  LIS_INT i;
  LIS_SCALAR t;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      z[i] = x[i]*y[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_pdiv"
LIS_INT lis_array_pdiv(LIS_INT n, LIS_SCALAR *x, LIS_SCALAR *y, LIS_SCALAR *z)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      z[i] = x[i]/y[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_set_all"
LIS_INT lis_array_set_all(LIS_INT n, LIS_SCALAR alpha, LIS_SCALAR *v)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      v[i] = alpha; 
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_abs"
LIS_INT lis_array_abs(LIS_INT n, LIS_SCALAR *v)
{
  LIS_INT i;
  
  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      v[i] = abs(v[i]);
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_reciprocal"
LIS_INT lis_array_reciprocal(LIS_INT n, LIS_SCALAR *v)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      v[i] = 1/v[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_shift"
LIS_INT lis_array_shift(LIS_INT n, LIS_SCALAR t, LIS_SCALAR *v)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  for(i=0;i<n;i++)
    {
      v[i] = v[i]+t;
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_dot"
LIS_INT lis_array_dot(LIS_INT n, LIS_SCALAR *v, LIS_SCALAR *dot)
{
	LIS_INT i;
	LIS_SCALAR t;

	LIS_DEBUG_FUNC_IN;

	t = 0.0;
	for(i=0;i<n;i++)
	{
		t += v[i]*v[i];
	}
	*dot = t;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_dot2"
LIS_INT lis_array_dot2(LIS_INT n, LIS_SCALAR *x, LIS_SCALAR *y, LIS_SCALAR *alpha)
{
  LIS_INT i;

  LIS_DEBUG_FUNC_IN;

  *alpha = 0;
  for(i=0;i<n;i++)
    {
      *alpha = *alpha + x[i] * y[i];
    }

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_nrm2"
LIS_INT lis_array_nrm2(LIS_INT n, LIS_SCALAR *v, LIS_SCALAR *nrm2)
{
	LIS_INT i;
	LIS_SCALAR t;

	LIS_DEBUG_FUNC_IN;

	t = 0.0;
	for(i=0;i<n;i++)
	{
		t += v[i]*v[i];
	}
	*nrm2 = sqrt(t);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_nrm1"
LIS_INT lis_array_nrm1(LIS_INT n, LIS_SCALAR *v, LIS_SCALAR *nrm1)
{
	LIS_INT i;
	LIS_SCALAR t;

	LIS_DEBUG_FUNC_IN;

	t = 0.0;
	for(i=0;i<n;i++)
	{
		t += fabs(v[i]);
	}
	*nrm1 = t;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_sum"
LIS_INT lis_array_sum(LIS_INT n, LIS_SCALAR *v, LIS_SCALAR *sum)
{
	LIS_INT i;
	LIS_SCALAR t;

	LIS_DEBUG_FUNC_IN;

	t = 0.0;
	for(i=0;i<n;i++)
	{
		t += v[i];
	}
	*sum = t;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_matvec"
LIS_INT lis_array_matvec(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *b, LIS_SCALAR *c, LIS_INT op)
{
	LIS_INT			i,j;
	LIS_SCALAR	t;

	LIS_DEBUG_FUNC_IN;

	/* c = A*b */

	if( op==LIS_INS_VALUE )
	{
		switch(n)
		{
		case 1:
			c[0] = a[0]*b[0];
			break;
		case 2:
			c[0] = a[0]*b[0] + a[2]*b[1];
			c[1] = a[1]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] = a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] = a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] = a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[j*n+i] * b[j];
				}
				c[i] = t;
			}
			break;
		}
	}
	else if( op==LIS_SUB_VALUE )
	{
		switch(n)
		{
		case 1:
			c[0] -= a[0]*b[0];
			break;
		case 2:
			c[0] -= a[0]*b[0] + a[2]*b[1];
			c[1] -= a[1]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] -= a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] -= a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] -= a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[j*n+i] * b[j];
				}
				c[i] -= t;
			}
			break;
		}
	}
	else
	{
		switch(n)
		{
		case 1:
			c[0] += a[0]*b[0];
			break;
		case 2:
			c[0] += a[0]*b[0] + a[2]*b[1];
			c[1] += a[1]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] += a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] += a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] += a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[j*n+i] * b[j];
				}
				c[i] += t;
			}
			break;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_matvect"
LIS_INT lis_array_matvect(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *b, LIS_SCALAR *c, LIS_INT op)
{
	LIS_INT			i,j;
	LIS_SCALAR	t;

	LIS_DEBUG_FUNC_IN;

	/* c = A*b */

	if( op==LIS_INS_VALUE )
	{
		switch(n)
		{
		case 1:
			c[0] = a[0]*b[0];
			break;
		case 2:
			c[0] = a[0]*b[0] + a[1]*b[1];
			c[1] = a[2]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
			c[1] = a[3]*b[0] + a[4]*b[1] + a[5]*b[2];
			c[2] = a[6]*b[0] + a[7]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[i*n+j] * b[j];
				}
				c[i] = t;
			}
			break;
		}
	}
	else if( op==LIS_SUB_VALUE )
	{
		switch(n)
		{
		case 1:
			c[0] -= a[0]*b[0];
			break;
		case 2:
			c[0] -= a[0]*b[0] + a[1]*b[1];
			c[1] -= a[2]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] -= a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
			c[1] -= a[3]*b[0] + a[4]*b[1] + a[5]*b[2];
			c[2] -= a[6]*b[0] + a[7]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[i*n+j] * b[j];
				}
				c[i] -= t;
			}
			break;
		}
	}
	else
	{
		switch(n)
		{
		case 1:
			c[0] += a[0]*b[0];
			break;
		case 2:
			c[0] += a[0]*b[0] + a[1]*b[1];
			c[1] += a[2]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] += a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
			c[1] += a[3]*b[0] + a[4]*b[1] + a[5]*b[2];
			c[2] += a[6]*b[0] + a[7]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[i*n+j] * b[j];
				}
				c[i] += t;
			}
			break;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_matvec2"
LIS_INT lis_array_matvec2(LIS_INT m, LIS_INT n, LIS_SCALAR *a, LIS_INT lda, LIS_SCALAR *b, LIS_SCALAR *c, LIS_INT op)
{
	LIS_INT			i,j;
	LIS_SCALAR	t;

	LIS_DEBUG_FUNC_IN;

	/* c = A*b */

	if( op==LIS_INS_VALUE )
	{
		for(i=0;i<m;i++)
		{
			t = 0.0;
			for(j=0;j<n;j++)
			{
				t += a[j*lda+i] * b[j];
			}
			c[i] = t;
		}
	}
	else if( op==LIS_SUB_VALUE )
	{
		for(i=0;i<m;i++)
		{
			t = 0.0;
			for(j=0;j<n;j++)
			{
				t += a[j*lda+i] * b[j];
			}
			c[i] -= t;
		}
	}
	else if( op==LIS_ADD_VALUE )
	{
		for(i=0;i<m;i++)
		{
			t = 0.0;
			for(j=0;j<n;j++)
			{
				t += a[j*lda+i] * b[j];
			}
			c[i] += t;
		}
	}
	else
	{
		switch(n)
		{
		case 1:
			c[0] += a[0]*b[0];
			break;
		case 2:
			c[0] += a[0]*b[0] + a[2]*b[1];
			c[1] += a[1]*b[0] + a[3]*b[1];
			break;
		case 3:
			c[0] += a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] += a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] += a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			break;
		default:
			for(i=0;i<n;i++)
			{
				t = 0.0;
				for(j=0;j<n;j++)
				{
					t += a[j*n+i] * b[j];
				}
				c[i] += t;
			}
			break;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_matinv"
LIS_INT lis_array_matinv(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *b, LIS_SCALAR *c)
{
	LIS_INT i,j,k;
	LIS_SCALAR t;

	LIS_DEBUG_FUNC_IN;

	for(i=0;i<n;i++)
	{
		c[i] = -b[i] * a[0];
		for(j=1;j<n;j++)
		{
			t = -b[j*n+i];
			for(k=0;k<j-1;k++)
			{
				t -= c[k*n+i] * a[j*n+k];
			}
			c[j*n+i] = t * a[j*n+j];
		}
	}
	for(i=0;i<n;i++)
	{
		for(j=n-1;j>=0;j--)
		{
			t = c[j*n+i];
			for(k=j+1;k<n;k++)
			{
				t -= c[k*n+i] * a[j*n+k];
			}
			c[j*n+i] = t;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_LUdecomp"
LIS_INT lis_array_LUdecomp(LIS_INT n, LIS_SCALAR *a)
{
	LIS_INT i,j,k;
	LIS_SCALAR t;

	LIS_DEBUG_FUNC_IN;

	for(k=0;k<n;k++)
	{
		a[k*n+k] = 1.0 / a[k*n+k];
		for(i=k+1;i<n;i++)
		{
			t = a[k*n+i] * a[k*n+k];
			for(j=k+1;j<n;j++)
			{
				a[j*n+i] -= t * a[j*n+k];
			}
			a[k*n+i] = t;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_invGauss"
LIS_INT lis_array_invGauss(LIS_INT n, LIS_SCALAR *a)
{
	LIS_INT i,j,k;
	LIS_SCALAR t,*lu;

	LIS_DEBUG_FUNC_IN;

	lu = (LIS_SCALAR *)lis_malloc(n*n*sizeof(LIS_SCALAR), "lis_array_invGauss::lu");
	memcpy(lu,a,n*n*sizeof(LIS_SCALAR));
	for(k=0;k<n;k++)
	{
		lu[k*n+k] = 1.0 / lu[k*n+k];
		for(i=k+1;i<n;i++)
		{
			t = lu[k*n+i] * lu[k*n+k];
			for(j=k+1;j<n;j++)
			{
				lu[j*n+i] -= t * lu[j*n+k];
			}
			lu[k*n+i] = t;
		}
	}
	for(k=0;k<n;k++)
	{
		for(i=0;i<n;i++)
		{
			 t = (i==k);
			 for(j=0;j<i;j++)
			 {
				 t -= lu[j*n+i] * a[k*n+j];
			 }
			 a[k*n+i] = t;
		}
		for(i=n-1;i>=0;i--)
		{
			t = a[k*n+i];
			for(j=i+1;j<n;j++)
			{
				t -= lu[j*n+i] * a[k*n+j];
			}
			a[k*n+i] = t * lu[i*n+i];
		}
	}
	free(lu);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_invvec"
LIS_INT lis_array_invvec(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *x, LIS_SCALAR *y)
{

	LIS_DEBUG_FUNC_IN;

	/* y = inv(a) * x */
	LIS_INT i,j;
	LIS_SCALAR t;

	for(i=0;i<n;i++)
	{
		t = x[i];
		for(j=0;j<i;j++)
		{
			t -= a[j*n+i] * y[j];
		}
		y[i] = t;
	}
	for(i=n-1;i>=0;i--)
	{
		t = y[i];
		for(j=i+1;j<n;j++)
		{
			t -= a[j*n+i] * y[j];
		}
		y[i] = t * a[i*n+i];
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_invvect"
LIS_INT lis_array_invvect(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *x, LIS_SCALAR *y)
{

	LIS_DEBUG_FUNC_IN;

	/* y = inv(a) * x */
	LIS_INT i,j;

	for(i=0;i<n;i++) y[i] = x[i];
	for(i=0;i<n;i++)
	{
		y[i] = a[i*n+i] * y[i];
		for(j=i+1;j<n;j++)
		{
			y[j] -= a[j*n+i] * y[i];
		}
	}
	for(i=n-1;i>=0;i--)
	{
		for(j=0;j<i;j++)
		{
			y[j] -= a[j*n+i] * y[i];
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_matmat"
LIS_INT lis_array_matmat(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *b, LIS_SCALAR *c, LIS_INT op)
{
	LIS_INT			i,j,l;

	LIS_DEBUG_FUNC_IN;

	/* C = A*B */

	if( op==LIS_INS_VALUE )
	{
		switch(n)
		{
		case 1:
			c[0] = a[0]*b[0];
			break;
		case 2:
			c[0] = a[0]*b[0] + a[2]*b[1];
			c[1] = a[1]*b[0] + a[3]*b[1];
			c[2] = a[0]*b[2] + a[2]*b[3];
			c[3] = a[1]*b[2] + a[3]*b[3];
			break;
		case 3:
			c[0] = a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] = a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] = a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			c[3] = a[0]*b[3] + a[3]*b[4] + a[6]*b[5];
			c[4] = a[1]*b[3] + a[4]*b[4] + a[7]*b[5];
			c[5] = a[2]*b[3] + a[5]*b[4] + a[8]*b[5];
			c[6] = a[0]*b[6] + a[3]*b[7] + a[6]*b[8];
			c[7] = a[1]*b[6] + a[4]*b[7] + a[7]*b[8];
			c[8] = a[2]*b[6] + a[5]*b[7] + a[8]*b[8];
			break;
		default:
			for(j=0;j<n;j++)
			  {
			    for(i=0;i<n;i++)
			      {
				c[j*n+i] = 0.0;
			      }
			    for(l=0;l<n;l++)
			      {
				for(i=0;i<n;i++)
				  {
				    c[j*n+i] += b[j*n+l] * a[l*n+i];
				  }
			      }
			  }
			break;
		}
	}
	else if( op==LIS_SUB_VALUE )
	{
		switch(n)
		{
		case 1:
			c[0] -= a[0]*b[0];
			break;
		case 2:
			c[0] -= a[0]*b[0] + a[2]*b[1];
			c[1] -= a[1]*b[0] + a[3]*b[1];
			c[2] -= a[0]*b[2] + a[2]*b[3];
			c[3] -= a[1]*b[2] + a[3]*b[3];
			break;
		case 3:
			c[0] -= a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] -= a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] -= a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			c[3] -= a[0]*b[3] + a[3]*b[4] + a[6]*b[5];
			c[4] -= a[1]*b[3] + a[4]*b[4] + a[7]*b[5];
			c[5] -= a[2]*b[3] + a[5]*b[4] + a[8]*b[5];
			c[6] -= a[0]*b[6] + a[3]*b[7] + a[6]*b[8];
			c[7] -= a[1]*b[6] + a[4]*b[7] + a[7]*b[8];
			c[8] -= a[2]*b[6] + a[5]*b[7] + a[8]*b[8];
			break;
		default:
			for(j=0;j<n;j++)
			{
				for(l=0;l<n;l++)
				{
					for(i=0;i<n;i++)
					{
						c[j*n+i] -= b[j*n+l] * a[l*n+i];
					}
				}
			}
			break;
		}
	}
	else
	{
		switch(n)
		{
		case 1:
			c[0] += a[0]*b[0];
			break;
		case 2:
			c[0] += a[0]*b[0] + a[2]*b[1];
			c[1] += a[1]*b[0] + a[3]*b[1];
			c[2] += a[0]*b[2] + a[2]*b[3];
			c[3] += a[1]*b[2] + a[3]*b[3];
			break;
		case 3:
			c[0] += a[0]*b[0] + a[3]*b[1] + a[6]*b[2];
			c[1] += a[1]*b[0] + a[4]*b[1] + a[7]*b[2];
			c[2] += a[2]*b[0] + a[5]*b[1] + a[8]*b[2];
			c[3] += a[0]*b[3] + a[3]*b[4] + a[6]*b[5];
			c[4] += a[1]*b[3] + a[4]*b[4] + a[7]*b[5];
			c[5] += a[2]*b[3] + a[5]*b[4] + a[8]*b[5];
			c[6] += a[0]*b[6] + a[3]*b[7] + a[6]*b[8];
			c[7] += a[1]*b[6] + a[4]*b[7] + a[7]*b[8];
			c[8] += a[2]*b[6] + a[5]*b[7] + a[8]*b[8];
			break;
		default:
			for(j=0;j<n;j++)
			{
				for(l=0;l<n;l++)
				{
					for(i=0;i<n;i++)
			        	{
						c[j*n+i] += b[j*n+l] * a[l*n+i];
					}
				}
			}
			break;
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_matmat2"
LIS_INT lis_array_matmat2(LIS_INT m, LIS_INT n, LIS_INT k, LIS_SCALAR *a, LIS_INT lda, LIS_SCALAR *b, LIS_INT ldb, LIS_SCALAR *c, LIS_INT ldc, LIS_INT op)
{
	LIS_INT			i,j,l;

	LIS_DEBUG_FUNC_IN;

	/* C = A*B */

	if( op==LIS_INS_VALUE )
	{
		for(j=0;j<n;j++)
		{
			for(i=0;i<m;i++)
			{
				c[j*ldc+i] = 0.0;
			}
			for(l=0;l<k;l++)
			{
				for(i=0;i<m;i++)
				{
					c[j*ldc+i] += b[j*ldb+l] * a[l*lda+i];
				}
			}
		}
	}
	else if( op==LIS_SUB_VALUE )
	{
		for(j=0;j<n;j++)
		{
			for(l=0;l<k;l++)
			{
				for(i=0;i<m;i++)
				{
					c[j*ldc+i] -= b[j*ldb+l] * a[l*lda+i];
				}
			}
		}
	}
	else
	{
		for(j=0;j<n;j++)
		{
			for(l=0;l<k;l++)
			{
				for(i=0;i<m;i++)
				{
					c[j*ldc+i] += b[j*ldb+l] * a[l*lda+i];
				}
			}
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_solve"
LIS_INT lis_array_solve(LIS_INT n, LIS_SCALAR *aa, LIS_SCALAR *b, LIS_SCALAR *x,
LIS_SCALAR *a)
{
	LIS_INT i,j,k,imax,swap;
	LIS_SCALAR t,al;

	LIS_DEBUG_FUNC_IN;

	for(i=0;i<n*n;i++) a[i] = aa[i];

	switch( n )
	{
	case 1:
		x[0] = b[0] / a[0];
		break;
	case 2:
		a[0]  = 1.0 / a[0];
		a[1] *= a[0];
		a[3] -= a[1] * a[2];
		a[3]  = 1.0 / a[3];
		/* forward sub */
		x[0] = b[0];
		x[1] = b[1] - a[1] * x[0];
		/* backward sub */
		x[1] *= a[3];
		x[0] -= a[2] * x[1];
		x[0] *= a[0];
		break;
	default:
		for(k=0;k<n;k++)
		{
			a[k*n+k] = 1.0 / a[k*n+k];
			for(i=k+1;i<n;i++)
			{
				t = a[k*n+i] * a[k*n+k];
				for(j=k+1;j<n;j++)
				{
					a[j*n+i] -= t * a[j*n+k];
				}
				a[k*n+i] = t;
			}
		}

		/* forward sub */
		for(i=0;i<n;i++)
		{
			x[i] = b[i];
			for(j=0;j<i;j++)
			{
				x[i] -= a[j*n+i] * x[j];
			}
		}
		/* backward sub */
		for(i=n-1;i>=0;i--)
		{
			for(j=i+1;j<n;j++)
			{
				x[i] -= a[j*n+i] * x[j];
			}
			x[i] *= a[i*n+i];
		}
		break;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_cgs"
LIS_INT lis_array_cgs(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *q, LIS_SCALAR *r)
{
  LIS_INT i, j, k; 
  LIS_SCALAR *a_k, nrm2;
  LIS_REAL tol;

  LIS_DEBUG_FUNC_IN;

  tol = 1e-12;

  a_k = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR), "lis_array_cgs::a_k");

  for (i=0;i<n*n;i++)
    {
    q[i] = 0.0;
    r[i] = 0.0;
    }

  for (k=0;k<n;k++)
    {
      for (i=0;i<n;i++)
	{
	  a_k[i] = a[i*n+k];
	}
      for (j=0;j<k;j++)
	{
	  r[j*n+k] = 0; 
	  for (i=0;i<n;i++)
	    {
	      r[j*n+k] += q[i*n+j] * a[i*n+k];
	    }
	  for (i=0;i<n;i++)
	    {
	      a_k[i] -= r[j*n+k] * q[i*n+j];
	    }
	}
      lis_array_nrm2(n, &a_k[0], &nrm2);
      r[k*n+k] = nrm2;
      if (nrm2<tol) break; 
      for (i=0;i<n;i++)
	{
	  q[i*n+k] = a_k[i] / nrm2;
	}
      
    }

  lis_free(a_k);


  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
} 

#undef __FUNC__
#define __FUNC__ "lis_array_mgs"
LIS_INT lis_array_mgs(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *q, LIS_SCALAR *r)
{
  LIS_INT i, j, k; 
  LIS_SCALAR *a_j, nrm2;
  LIS_REAL tol;

  LIS_DEBUG_FUNC_IN;

  tol = 1e-12;

  a_j = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR), "lis_array_mgs::a_j");

  for (j=0;j<n;j++)
    {
      for (i=0;i<n;i++)
	{
	  a_j[i] = a[i*n+j];
	}
      lis_array_nrm2(n, &a_j[0], &nrm2);
      r[j*n+j] = nrm2;
      for (i=0;i<n;i++)
	{
	  if (nrm2<tol) break; 
	  q[i*n+j] = a_j[i] / nrm2;
	}
      for (k=j+1;k<n;k++)
	{
	  r[j*n+k] = 0; 
	  for (i=0;i<n;i++)
	    {
	      r[j*n+k] += q[i*n+j] * a[i*n+k];
	    }
	  for (i=0;i<n;i++)
	    {
	      a[i*n+k] -= r[j*n+k] * q[i*n+j];
	    }
	}
    }
  lis_free(a_j);

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_qr"
LIS_INT lis_array_qr(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *q, LIS_SCALAR *r)
{
  LIS_INT i, j, k, iter, maxiter; 
  LIS_SCALAR *a0, a_tmp;
  LIS_REAL err, err0, tol;

  LIS_DEBUG_FUNC_IN;

  maxiter = 100000;
  tol = 1e-12;

  a0 = (LIS_SCALAR *)lis_malloc(n*n*sizeof(LIS_SCALAR), "lis_array_qr::a0");
  iter = 0;
  while (iter < maxiter)
    {
      iter = iter + 1;
      lis_array_cgs(n, a, q, r);
      for (i=0;i<n;i++)
	{
	  for (j=0;j<n;j++)
	    {
	      a[i*n+j] = 0;
	      for (k=0;k<n;k++)
		{
		  a[i*n+j] += r[i*n+k] * q[k*n+j];
		}
	    }
	}
      err = sqrt(a[n*n-2] * a[n*n-2]); 
      if (err<tol) break;
    }

  lis_free(a0);

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_array_power"
LIS_INT lis_array_power(LIS_INT n, LIS_SCALAR *a, LIS_SCALAR *x, LIS_SCALAR *mu, LIS_INT maxiter, LIS_REAL tol, LIS_REAL *err)
{
  LIS_INT               i, iter;
  LIS_REAL          nrm2;
  LIS_SCALAR        *z,*q;
  double            time, ptime;

  LIS_DEBUG_FUNC_IN;

  iter=0;
  lis_array_set_all(n, 1.0, x);
  z = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR), "lis_array_power::z");

  q = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR), "lis_array_power::q");
  time = lis_wtime();
  while (iter<maxiter)
    {
      iter = iter+1;
      lis_array_nrm2(n, x, &nrm2);
      lis_array_scale(n, 1/nrm2, x);
      lis_array_matvec(n, a, x, z, LIS_INS_VALUE);
      lis_array_dot2(n, x, z, mu); 
      lis_array_axpyz(n, -(*mu),x,z,q); 
      lis_array_nrm2(n, q, err); 
      *err = fabs(*err / (*mu));
      if (*err<tol) break;  
      lis_array_copy(n, z, x);
    }
  lis_free(z);
  lis_free(q);

  LIS_DEBUG_FUNC_OUT;
  return LIS_SUCCESS;
}




